/*
** in.mfingerd: finger daemon which lists users on several machines
**
** Copyright (C) 1997 Mark Rigby-Jones
** mrj@ox.compsoc.org.uk <*> http://www.ox.compsoc.org.uk/~mrj/
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** If you do not have a copy of the GNU General Public License then
** you can get one by writing to the Free Software Foundation, Inc.,
** 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdlib.h>
#include <stdio.h>
#include <strings.h>
#include <unistd.h>

#define VERSION  "in.mfingerd v3.2 by Mark Rigby-Jones, 26th Feb 1997"
#define CONFIG_DIR "/usr/local/etc"
#define BUFLEN 80

void substr(char *dest, char *src, int from, int size) {
  int i;
  for (i = 0; i < size; i++) {
    dest[i] = src[i + from];
  }
  dest[size] = 0;
}

void main(int argc, char** argv) {
  char user[BUFLEN], buf[BUFLEN], finger[BUFLEN + 7];
  char host[BUFLEN], name[BUFLEN], hostname[BUFLEN];
  char config_file[sizeof(CONFIG_DIR) + 15];
  char config_info[sizeof(CONFIG_DIR) + 15];
  char input_file[25], output_file[25];
  char login[30], tty[5], idle[6], time[13];
  char *change;
  FILE *config, *info, *input, *output;
  int all, clients, mfinger, free, help, site, local, version;
  int found = 0;
  pid_t pid;

  gethostname(hostname, BUFLEN);
  pid = getpid();
  sprintf(input_file, "/tmp/mfinger.%d.in", pid);
  sprintf(output_file, "/tmp/mfinger.%d.out", pid);

  fgets(user, BUFLEN, stdin);
  change = strchr(user, '\n');
  if (change != NULL) {
    change[0] = 0;
  }
  change = strchr(user, '\r');
  if (change != NULL) {
    change[0] = 0;
  }

  if (strchr(user, '@') != NULL) {
    printf("%s: Finger forwarding not permitted!\n", argv[0]);
    exit(1);
  }
  if ((strpbrk(user, "|;&()<>`") != NULL)) {
    printf("%s: Illegal character in username!\n", argv[0]);
    exit(1);
  }

  if (strncmp(user, "/W", 2) == 0) {
    if (user[3] == 0) {
      user[0] = 0;
    } else {
      user[0] = '-';
      user[1] = 'l';
    }
  }

  sprintf(config_file, "%s/mfingerd.conf", CONFIG_DIR);
  config = fopen(config_file, "r");
  if (config == NULL) {
    printf("%s: Unable to open configuration!\n", argv[0]);
    exit(system("finger $user"));
  }

  all = strcmp(user, "");
  clients = strcmp(user, ".clients");
  mfinger = strcmp(user, ".mfinger");
  free = strcmp(user, ".free");
  help = strcmp(user, ".help");
  site = strcmp(user, ".site");
  local = strcmp(user, ".local");
  version = strcmp(user, ".version");

  if ((all == 0) || (clients == 0) || (mfinger == 0)) {
    fgets(buf, BUFLEN, config);
    if (all == 0) {
      while ((buf != NULL) && (strncmp(buf, "#HOSTS", 6) != 0)) {
	if (strncmp(buf, "#MOTD", 5) == 0) {
	  fgets(buf, BUFLEN, config);
	  while (buf[0] != '#') {
	    printf("%s", buf);
	    fgets(buf, BUFLEN, config);
	  }
	} else {
	  fgets(buf, BUFLEN, config);
	}
      }
      output = fopen(output_file, "w");
    } else {
      while ((buf !=NULL) && (strncmp(buf, "#HOSTS", 6) != 0)) {
	fgets(buf, BUFLEN, config);
      }
    }
    if (clients == 0) {
      printf("%s: Client machines checked from this host:\n", argv[0]);
    }
    while (fgets(buf, BUFLEN, config)) {
      if (buf[0] != '#') {
	change = strchr(buf, '\n');
	if (change != NULL) {
	  change[0] = 0;
	}
	change = strchr(buf, '\t');
	if (change == NULL) {
	  strcpy(host, buf);
	} else {
	  strcpy(name, (change + 1));
	  change[0] = 0;
	  strcpy(host, buf);
	}
	if (strcmp(host, "localhost") == 0) {
	  strcpy(host, hostname);
	}
	if (change == NULL) {
	  strcpy(name, host);
	}
	if (clients == 0) {
	  printf("%25s - %s\n", name, host);
	} else if (strcmp(host, hostname) == 0) {
          sprintf(finger, "finger > %s", input_file);
	  system(finger);
	  input = fopen(input_file, "r");
	  fgets(buf, BUFLEN, input);
	  while (fgets(buf, BUFLEN, input) && (strcmp(buf, "\n") != 0)) {
	    substr(login, buf, 0, 29);
	    substr(tty, buf, 30, 4);
	    substr(idle, buf, 35, 5);
	    substr(time, buf, 41, 12);
	    if (all == 0) {
	      fprintf(output, "%s %s %s %s %s\n", login, time, idle, tty, name);
	      found++;
	    } else {
	      printf("%s %s %s %s %s\n", login, time, idle, tty, name);
	    }
	  }
	  fclose(input);
	  unlink(input_file);
	} else if (all == 0) {
	  sprintf(finger, "finger .mfinger@%s > %s", host, input_file);
	  system(finger);
	  input = fopen(input_file, "r");
	  fgets(buf, BUFLEN, input);
	  while (fgets(buf, BUFLEN, input) && (strcmp(buf, "\n") != 0)) {
	    substr(finger, buf, 0, 53);
	    fprintf(output, "%s %s\n", finger, name);
	    found++;
	  }
	  fclose(input);
	  unlink(input_file);
	}
      }
    }
    if (all == 0) {
      fclose(output);
      if (found == 0) {
	unlink(output_file);
	printf("Nobody logged in!\n");
      } else {
	sprintf(finger, "sort %s > %s", output_file, input_file);
	system(finger);
	unlink(output_file);
	printf("Login    Real name            Login time   Idle  tty  Where\n");
	input = fopen(input_file, "r");
	while (fgets(buf, BUFLEN, input)) {      
	  printf("%s", buf);
	}
	fclose(input);
	unlink(input_file);
      }
    }
  } else if (free == 0) {
    printf("%s: Free machines?  Surely that's meaningless on a multiuser OS!  ;-)\n", argv[0]);
  } else if ((help == 0) || (site == 0)) {
    sprintf(config_info, "%s/mfingerd%s", CONFIG_DIR, user);
    info = fopen(config_info, "r");
    if (info != NULL) {
      while (fgets(buf, BUFLEN, info) != NULL) {
	printf("%s", buf);
      }
      fclose(info);
    } else {
      printf("%s: No %s information available.\n", argv[0], user);
    }
  } else if (local == 0) {
    system("finger");
  } else if (version == 0) {
    printf("%s\n", VERSION);
  } else if (argc == 1) {
    sprintf(finger, "finger %s", user);
    system(finger);
  } else {
    sprintf(finger, "finger %s@%s", user, argv[1]);
    system(finger);
  }
  fclose(config);
}
